import pygame
import numpy as np
import math
from pygame.math import Vector2
import os
from scipy.spatial import ConvexHull
import random
from track_gen import gen_track
import copy
from pygame.color import THECOLORS

# Fast SLAM covariance
Q = np.diag([3.0, np.deg2rad(10.0)])**2
R = np.diag([1.0, np.deg2rad(20.0)])**2

#  Simulation parameter
Qsim = np.diag([0.5, np.deg2rad(2.0)])**2
Rsim = np.diag([0.5, np.deg2rad(10.0)])**2
OFFSET_YAWRATE_NOISE = 0.1

DT = 0.1
MAX_RANGE = 30  # maximum observation range
M_DIST_TH = 5.0  # Threshold of Mahalanobis distance for data association.
STATE_SIZE = 3  # State size [x,y,yaw]
LM_SIZE = 2  # LM srate size [x,y]
N_PARTICLE = 100  # number of particle
NTH = N_PARTICLE / 1.5  # Number of particle for re-sampling
RFID = []


pygame.display.set_caption("Simple Sim")
WIDTH = 1280
HEIGHT = 720
SCREEN = pygame.display.set_mode((WIDTH, HEIGHT))
CLOCK = pygame.time.Clock()
TICKS = 60
EXIT = False
CENTER = (WIDTH/2, HEIGHT/2)
GOAL_ONE = (0,0)


#car stuff
LENGTH = 42
brake_deceleration = 50
free_deceleration = 25
max_acceleration = 5.0
max_steering = 30
#this is the scaling factor between the game and the real world for physics
PPU = 3
STEERING = 0.0
ACCELERATION = 0.0
VELOCITY = Vector2(0.0,0.0)
carPos = Vector2(0,0)

SCORE = 0
LEVEL = 0
mids = np.zeros((STATE_SIZE, 1))

class Particle:

		def __init__(self, N_LM):
			self.w = 1.0 / N_PARTICLE
			self.x = 0.0
			self.y = 0.0
			self.yaw = 0.0
			#self.Velocity = VELOCITY
			self.P = np.eye(3)
			# landmark x-y positions
			self.lm = np.zeros((N_LM, LM_SIZE))
			# landmark position covariance
			self.lmP = np.zeros((N_LM * LM_SIZE, LM_SIZE))

class SimState:

	def __init__(self):
		global VELOCITY
		global DT
		global mids
		global GOAL_ONE
		global RFID
		global carPos
		global N_LM

		pygame.init()
		print(__file__ + " start!!")

		current_dir = os.path.dirname(os.path.abspath(__file__))
	   
		car_image_path = os.path.join(current_dir, "car.png")
		self.car_image = pygame.image.load(car_image_path)
		self.car_image = pygame.transform.scale(self.car_image, (20, 10))
			
		wheel_image_path = os.path.join(current_dir, "wheel.png")
		self.wheel_image = pygame.image.load(wheel_image_path)
		self.wheel_image = pygame.transform.scale(self.wheel_image, (100,100))

		self.time = 0.0

		'''self.yellowcones = np.array(self.outer_track_gen())
		self.mids = np.array(self.gen_mids(self.yellowcones))
		self.bluecones = np.array(self.inner_track_gen(self.yellowcones))'''
		self.yellowcones,self.bluecones,self.mids = gen_track()

		for i in range(0,len(self.mids)-1):
			self.mids[i] = (self.move_to_mid(self.mids[i]))
		for i in range(0,len(self.yellowcones)-1):
			self.yellowcones[i] = (self.move_to_mid(self.yellowcones[i]))

		for i in range(0,len(self.bluecones)-1):
			self.bluecones[i] = (self.move_to_mid(self.bluecones[i]))

		#Move whole track so car is at first midpoint
		#Track is moved rather than car so SLAM can start at 0,0 always
		GOAL_ONE = copy.copy(self.mids[0])
		x = GOAL_ONE[0]
		y = GOAL_ONE[1]

		for i in range(0,len(self.mids)-1):
			self.mids[i] = (self.move_to_car(self.mids[i]))
		for i in range(0,len(self.yellowcones)-1):
			self.yellowcones[i] = (self.move_to_car(self.yellowcones[i]))

		for i in range(0,len(self.bluecones)-1):
			self.bluecones[i] = (self.move_to_car(self.bluecones[i]))

		

		RFID = np.concatenate((self.yellowcones,self.bluecones))

		N_LM = RFID.shape[0]


		# State Vector [x y yaw v]'
		#xEst = np.array([[mids[0][0]], [mids[0][1]], [0]])  # SLAM estimation
		self.xEst = np.zeros((STATE_SIZE, 1))  # True state
		#xTrue = np.array([[mids[0][0]], [mids[0][1]], [0]])
		self.xTrue = np.zeros((STATE_SIZE, 1))
		#xDR = np.array([[mids[0][0]], [mids[0][1]], [0]])  # Dead reckoning
		self.xDR = np.zeros((STATE_SIZE, 1))

		# history
		self.hxEst = self.xEst
		self.hxTrue = self.xTrue
		self.hxDR = self.xTrue

		self.particles = [Particle(N_LM) for i in range(N_PARTICLE)]

	
	def frame_step(self, action):

		global VELOCITY
		global DT
		global mids
		global GOAL_ONE
		global RFID
		global carPos
		global N_LM

		#event queue
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				EXITexit = True

		DT = CLOCK.get_time() / 1000

		#formulate u
		u = self.calc_input(self.xTrue, action)

		self.xTrue, z, self.xDR, ud = self.observation(self.xTrue, self.xDR, u, RFID)
		#print(xTrue)
			
		self.particles = self.fast_slam2(self.particles, ud, z)

		self.xEst = self.calc_final_state(self.particles)

		x_state = self.xEst[0: STATE_SIZE]

		# store data history
		self.hxEst = np.hstack((self.hxEst, x_state))
		self.hxDR = np.hstack((self.hxDR, self.xDR))
		self.hxTrue = np.hstack((self.hxTrue, self.xTrue))
			
		#update Velocity 
		VELOCITY += (ACCELERATION * DT,0)



		#Blitting stuff

		SCREEN.fill((255, 255, 255))

		for x in self.yellowcones:
				pygame.draw.circle(SCREEN, (254,203,0), self.coordinateMapping((int(x[0]), int(x[1]))), 3)
		for x in self.bluecones:
				pygame.draw.circle(SCREEN, (0,102,204), self.coordinateMapping((int(x[0]), int(x[1]))), 3)
		'''col = 255
		for x in self.mids:
				pygame.draw.circle(SCREEN, (col,col,col), self.coordinateMapping((int(x[0]), int(x[1]))), 10)
				col -= 1
		pygame.draw.circle(SCREEN, (0,255,0), self.coordinateMapping((int(self.mids[0][0]), int(self.mids[0][1]))), 10)
		pygame.draw.circle(SCREEN, (0,220,0), self.coordinateMapping((int(self.mids[1][0]), int(self.mids[1][1]))), 10)
		pygame.draw.circle(SCREEN, (0,200,0), self.coordinateMapping((int(self.mids[2][0]), int(self.mids[2][1]))), 10)
		pygame.draw.circle(SCREEN, (0,180,0), self.coordinateMapping((int(self.mids[3][0]), int(self.mids[3][1]))), 10)
		pygame.draw.circle(SCREEN, (0,150,0), self.coordinateMapping((int(self.mids[4][0]), int(self.mids[4][1]))), 10)
		pygame.draw.circle(SCREEN, (0,120,0), self.coordinateMapping((int(self.mids[5][0]), int(self.mids[5][1]))), 10)
		pygame.draw.circle(SCREEN, (0,100,0), self.coordinateMapping((int(self.mids[6][0]), int(self.mids[6][1]))), 10)
		pygame.draw.circle(SCREEN, (0,80,0), self.coordinateMapping((int(self.mids[7][0]), int(self.mids[7][1]))), 10)
		pygame.draw.circle(SCREEN, (0,50,0), self.coordinateMapping((int(self.mids[8][0]), int(self.mids[8][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,0), self.coordinateMapping((int(self.mids[9][0]), int(self.mids[9][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,255), self.coordinateMapping((int(self.mids[10][0]), int(self.mids[10][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,220), self.coordinateMapping((int(self.mids[11][0]), int(self.mids[11][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,180), self.coordinateMapping((int(self.mids[12][0]), int(self.mids[12][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,150), self.coordinateMapping((int(self.mids[13][0]), int(self.mids[13][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,120), self.coordinateMapping((int(self.mids[14][0]), int(self.mids[14][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,100), self.coordinateMapping((int(self.mids[15][0]), int(self.mids[15][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,80), self.coordinateMapping((int(self.mids[16][0]), int(self.mids[16][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,50), self.coordinateMapping((int(self.mids[17][0]), int(self.mids[17][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,0), self.coordinateMapping((int(self.mids[18][0]), int(self.mids[18][1]))), 10)
		pygame.draw.circle(SCREEN, (255,0,0), self.coordinateMapping((int(self.mids[19][0]), int(self.mids[19][1]))), 10)
		pygame.draw.circle(SCREEN, (220,0,0), self.coordinateMapping((int(self.mids[20][0]), int(self.mids[20][1]))), 10)
		pygame.draw.circle(SCREEN, (200,0,0), self.coordinateMapping((int(self.mids[21][0]), int(self.mids[21][1]))), 10)
		pygame.draw.circle(SCREEN, (180,0,0), self.coordinateMapping((int(self.mids[22][0]), int(self.mids[22][1]))), 10)
		pygame.draw.circle(SCREEN, (150,0,0), self.coordinateMapping((int(self.mids[23][0]), int(self.mids[23][1]))), 10)
		pygame.draw.circle(SCREEN, (120,0,0), self.coordinateMapping((int(self.mids[24][0]), int(self.mids[24][1]))), 10)
		pygame.draw.circle(SCREEN, (100,0,0), self.coordinateMapping((int(self.mids[25][0]), int(self.mids[25][1]))), 10)
		pygame.draw.circle(SCREEN, (80,0,0), self.coordinateMapping((int(self.mids[26][0]), int(self.mids[26][1]))), 10)
		pygame.draw.circle(SCREEN, (50,0,0), self.coordinateMapping((int(self.mids[27][0]), int(self.mids[27][1]))), 10)
		pygame.draw.circle(SCREEN, (20,0,0), self.coordinateMapping((int(self.mids[28][0]), int(self.mids[28][1]))), 10)
		pygame.draw.circle(SCREEN, (0,0,0), self.coordinateMapping((int(self.mids[29][0]), int(self.mids[29][1]))), 10)'''


				
		#SLAM visualisation
		#Visualise the cone locations
		for i in range(N_PARTICLE):
			for x in self.particles[i].lm:
				pygame.draw.circle(SCREEN, (0,0,0), self.coordinateMapping((int(x[0]), int(x[1]))), 5, 2)
			
		#vis pose history (true, est, dead reckoning)
		for x in range(len(self.hxTrue) ,len(self.hxTrue[0]) -1):
			pygame.draw.circle(SCREEN, (0,0,0), self.coordinateMapping((int(self.hxTrue[0,x]), int(self.hxTrue[1,x]))), 3)
		for x in range(len(self.hxEst),len(self.hxEst[0]) -1):
			pygame.draw.circle(SCREEN, (0,0,255), self.coordinateMapping((int(self.hxEst[0,x]), int(self.hxEst[1,x]))), 3)
		for x in range(len(self.hxDR),len(self.hxDR[0]) -1):
			pygame.draw.circle(SCREEN, (255,0,255), self.coordinateMapping((int(self.hxDR[0,x]), int(self.hxDR[1,x]))), 3)

		#create a car sprite at the correct angle
		rotated_car = pygame.transform.rotate(self.car_image, math.degrees(self.xTrue[2]))
		car_rect = rotated_car.get_rect()
		#create a wheel sprite at the right angle
		rotated_wheel = pygame.transform.rotate(self.wheel_image, STEERING)
		wheel_rect = rotated_wheel.get_rect()
		#blit the sprites
		carPos = Vector2(self.xTrue[0,0], self.xTrue[1,0])
		slamPos = Vector2(x_state[0,0], x_state[1,0])

		reward = self.reward_function(carPos, self.mids)

		done = False
		if reward < -1000:
			done = True

		state = self.state()

		SCREEN.blit(rotated_car, self.coordinateMapping((carPos) - (car_rect.width / 2, car_rect.height / 2)))
		pygame.draw.circle(SCREEN, (255,0,0), self.coordinateMapping((int(carPos.x), int(carPos.y))), MAX_RANGE, 1)
		SCREEN.blit(rotated_wheel, (Vector2(WIDTH - 150,50) - (wheel_rect.width / 2, wheel_rect.height / 2)))            

		pygame.display.flip()
		#make sure clock moves forward within THIS loop
		CLOCK.tick(TICKS)

		return state, reward, done 


	def fast_slam2(self, particles, u, z):

		particles = self.predict_particles(particles, u)

		particles = self.update_with_observation(particles, z)

		particles = self.resampling(particles)

		return particles


	def normalize_weight(self, particles):

		sumw = sum([p.w for p in particles])

		try:
			for i in range(N_PARTICLE):
				particles[i].w /= sumw
		except ZeroDivisionError:
			for i in range(N_PARTICLE):
				particles[i].w = 1.0 / N_PARTICLE

			return particles

		return particles


	def calc_final_state(self, particles):

		xEst = np.zeros((STATE_SIZE, 1))

		particles = self.normalize_weight(particles)

		for i in range(N_PARTICLE):
			xEst[0, 0] += particles[i].w * particles[i].x
			xEst[1, 0] += particles[i].w * particles[i].y
			xEst[2, 0] += particles[i].w * particles[i].yaw

		xEst[2, 0] = self.pi_2_pi(xEst[2, 0])
		# print(xEst)

		return xEst #+ np.array([[mids[0][0]], [mids[0][1]], [0]])


	def predict_particles(self, particles, u):

		for i in range(N_PARTICLE):
			px = np.zeros((STATE_SIZE, 1))
			px[0, 0] = particles[i].x
			px[1, 0] = particles[i].y
			px[2, 0] = particles[i].yaw
			ud = u + (np.random.randn(1, 2) @ R).T  # add noise
			px = self.motion_model(px, ud)
			particles[i].x = px[0, 0]
			particles[i].y = px[1, 0]
			particles[i].yaw = px[2, 0]

		return particles


	def add_new_lm(self, particle, z, Q):

		r = z[0]
		b = z[1]
		lm_id = int(z[2])

		s = math.sin(self.pi_2_pi(particle.yaw + b))
		c = math.cos(self.pi_2_pi(particle.yaw + b))

		particle.lm[lm_id, 0] = particle.x + r * c
		particle.lm[lm_id, 1] = particle.y + r * s

		# covariance
		Gz = np.array([[c, -r * s],
					   [s, r * c]])

		particle.lmP[2 * lm_id:2 * lm_id + 2] = Gz @ Q @ Gz.T

		return particle


	def compute_jacobians(self, particle, xf, Pf, Q):
		dx = xf[0, 0] - particle.x
		dy = xf[1, 0] - particle.y 
		d2 = dx**2 + dy**2
		d = math.sqrt(d2)

		zp = np.array(
			[d, self.pi_2_pi(math.atan2(dy, dx) - particle.yaw)]).reshape(2, 1)

		Hv = np.array([[-dx / d, -dy / d, 0.0],
					   [dy / d2, -dx / d2, -1.0]])

		Hf = np.array([[dx / d, dy / d],
					   [-dy / d2, dx / d2]])

		Sf = Hf @ Pf @ Hf.T + Q
		
		return zp, Hv, Hf, Sf


	def update_KF_with_cholesky(self, xf, Pf, v, Q, Hf):
		PHt = Pf @ Hf.T
		S = Hf @ PHt + Q

		S = (S + S.T) * 0.5
		SChol = np.linalg.cholesky(S).T
		SCholInv = np.linalg.inv(SChol)
		W1 = PHt @ SCholInv
		W = W1 @ SCholInv.T

		x = xf + W @ v
		P = Pf - W1 @ W1.T

		return x, P


	def update_landmark(self, particle, z, Q):

		lm_id = int(z[2])
		xf = np.array(particle.lm[lm_id, :]).reshape(2, 1)
		Pf = np.array(particle.lmP[2 * lm_id:2 * lm_id + 2])

		zp, Hv, Hf, Sf = self.compute_jacobians(particle, xf, Pf, Q)

		dz = z[0:2].reshape(2, 1) - zp
		dz[1, 0] = self.pi_2_pi(dz[1, 0])

		xf, Pf = self.update_KF_with_cholesky(xf, Pf, dz, Q, Hf)

		particle.lm[lm_id, :] = xf.T
		particle.lmP[2 * lm_id:2 * lm_id + 2, :] = Pf

		return particle


	def compute_weight(self, particle, z, Q):

		lm_id = int(z[2])
		xf = np.array(particle.lm[lm_id, :]).reshape(2, 1)
		Pf = np.array(particle.lmP[2 * lm_id:2 * lm_id + 2])
		zp, Hv, Hf, Sf = self.compute_jacobians(particle, xf, Pf, Q)

		dz = z[0:2].reshape(2, 1) - zp
		dz[1, 0] = self.pi_2_pi(dz[1, 0])

		try:
			invS = np.linalg.inv(Sf)
		except np.linalg.linalg.LinAlgError:
			return 1.0

		num = math.exp(-0.5 * dz.T @ invS @ dz)
		den = 2.0 * math.pi * math.sqrt(np.linalg.det(Sf))

		w = num / den

		return w


	def proposal_sampling(self, particle, z, Q):

		lm_id = int(z[2])
		xf = particle.lm[lm_id, :].reshape(2, 1)
		Pf = particle.lmP[2 * lm_id:2 * lm_id + 2]
		# State
		x = np.array([particle.x, particle.y, particle.yaw]).reshape(3, 1)
		P = particle.P
		zp, Hv, Hf, Sf = self.compute_jacobians(particle, xf, Pf, Q)

		Sfi = np.linalg.inv(Sf)
		dz = z[0:2].reshape(2, 1) - zp
		dz[1] = self.pi_2_pi(dz[1])

		Pi = np.linalg.inv(P)

		particle.P = np.linalg.inv(Hv.T @ Sfi @ Hv + Pi)  # proposal covariance
		x += particle.P @ Hv.T @ Sfi @ dz  # proposal mean

		particle.x = x[0, 0]
		particle.y = x[1, 0]
		particle.yaw = x[2, 0]

		return particle


	def update_with_observation(self, particles, z):

		for iz in range(len(z[0, :])):
			lmid = int(z[2, iz])

			for ip in range(N_PARTICLE):
				# new landmark
				if abs(particles[ip].lm[lmid, 0]) <= 0.01:
					particles[ip] = self.add_new_lm(particles[ip], z[:, iz], Q)
				# known landmark
				else:
					w = self.compute_weight(particles[ip], z[:, iz], Q)
					particles[ip].w *= w

					particles[ip] = self.update_landmark(particles[ip], z[:, iz], Q)
					particles[ip] = self.proposal_sampling(particles[ip], z[:, iz], Q)

		return particles


	def resampling(self, particles):
		"""
		low variance re-sampling
		"""

		particles = self.normalize_weight(particles)

		pw = []
		for i in range(N_PARTICLE):
			pw.append(particles[i].w)

		pw = np.array(pw)

		Neff = 1.0 / (pw @ pw.T)  # Effective particle number
		#  print(Neff)

		if Neff < NTH:  # resampling
			wcum = np.cumsum(pw)
			base = np.cumsum(pw * 0.0 + 1 / N_PARTICLE) - 1 / N_PARTICLE
			resampleid = base + np.random.rand(base.shape[0]) / N_PARTICLE

			inds = []
			ind = 0
			for ip in range(N_PARTICLE):
				while ((ind < wcum.shape[0] - 1) and (resampleid[ip] > wcum[ind])):
					ind += 1
				inds.append(ind)

			tparticles = particles[:]
			for i in range(len(inds)):
				particles[i].x = tparticles[inds[i]].x
				particles[i].y = tparticles[inds[i]].y
				particles[i].yaw = tparticles[inds[i]].yaw
				particles[i].lm = tparticles[inds[i]].lm[:, :]
				particles[i].lmP = tparticles[inds[i]].lmP[:, :]
				particles[i].w = 1.0 / N_PARTICLE

		return particles


	def calc_input(self,car, action):

		global ACCELERATION
		global STEERING
		global VELOCITY
		global SCORE

		
		velocity = Vector2(VELOCITY.x, VELOCITY.y)

		if (action == 1) or (action == 6) or (action == 9):
			if velocity.x < 0:
				ACCELERATION = brake_deceleration
			else:
				ACCELERATION += 1 * DT
				SCORE += 10
		elif (action == 2) or (action == 7) or (action == 10):
			if velocity.x > 0:
				ACCELERATION = -brake_deceleration
			else:
				ACCELERATION -= 1 * DT
		elif (action == 3) or (action == 8) or (action == 11):
			if abs(velocity.x) > DT * brake_deceleration:
				ACCELERATION = -math.copysign(brake_deceleration, velocity.x)
			else:
				ACCELERATION = -velocity.x / DT
		else:
			if abs(velocity.x) > DT * free_deceleration:
				ACCELERATION = -math.copysign(free_deceleration, velocity.x)
			else:
				if DT != 0:
					ACCELERATION = -velocity.x / DT
		#update acceleration 
		ACCELERATION = max(-max_acceleration, min(ACCELERATION, max_acceleration))
						
		#ANGLE
		if (action == 4) or (action == 6) or (action == 7) or (action == 8):
			STEERING -= 120 * DT
		elif (action == 5) or (action == 9) or (action == 10) or (action == 11):
			STEERING += 120 * DT
		#update angle
		STEERING = max(-max_steering, min(STEERING, max_steering))

		u = np.array([ACCELERATION, STEERING]).reshape(2,1)

		return u


	def observation(self,xTrue, xd, u, RFID):

		# calc true state
		xTrue = self.motion_model(xTrue, u)

		# add noise to range observation
		z = np.zeros((3, 0))

		for i in range(len(RFID[:, 0])):

			dx = RFID[i, 0] - xTrue[0,0]
			dy = RFID[i, 1] - xTrue[1,0]  
			d = (math.sqrt(dx**2 + dy**2)) 
			#print(d)
			angle = self.pi_2_pi(math.atan2(dy, dx) - xTrue[2,0])
			#print(angle)
			if d <= MAX_RANGE:
				dn = d + np.random.randn() * Qsim[0, 0]  # add noise
				anglen = angle + np.random.randn() * Qsim[1, 1]  # add noise
				zi = np.array([dn, self.pi_2_pi(anglen), i]).reshape(3, 1)
				#print(zi)
				z = np.hstack((z, zi))

		# add noise to input
		ud1 = u[0] + np.random.randn() * Rsim[0, 0]
		ud2 = u[1] + np.random.randn() * Rsim[1, 1] + OFFSET_YAWRATE_NOISE
		ud = np.array([ud1, ud2])
		
		xd = self.motion_model(xd, ud)

		return xTrue, z, xd, ud


	def motion_model(self, x, u):
		global VELOCITY
		global PPU

		pos = Vector2(x[0,0], x[1,0]) / PPU
		velocity = Vector2(VELOCITY.x, VELOCITY.y)
		angle = x[2,0]
		steering = u[1,0]
		acceleration = u[0,0]
		
		velocity += (acceleration * DT, 0)
		
		if steering:
			turning_radius = LENGTH / math.tan(math.radians(steering))
			angular_velocity = velocity.x / turning_radius
		else:
			angular_velocity = 0

		pos += velocity.rotate(math.degrees(-angle)) * DT
		pos = pos * PPU
		angle += (angular_velocity) * DT

		x = np.array([pos.x , pos.y, self.pi_2_pi(angle)]).reshape(3,1)   

		return x


	def pi_2_pi(self, angle):
		return (angle + math.pi) % (2 * math.pi) - math.pi

	def dist(self, point1, point2):
		x1, y1 = point1[0], point1[1]
		x2, y2 = point2[0], point2[1]
		d = math.sqrt(math.pow(x2 - x1, 2) + math.pow(y2 - y1, 2))
		return d

	def reward_function(self, carPos, levels):
	    global SCORE
	    global LEVEL
	    global TICKS

	    in_track = False
	    
	    for i in range(len(levels) -1):
	        if (self.dist((carPos.x,carPos.y),levels[i]) <= 10):
	            in_track = True
	            if(LEVEL == (i)):
	                SCORE += ((len(levels)) - i ) #* int((1/(pygame.time.get_ticks()) * 10000)))
	                LEVEL = LEVEL + 1
	            else:
	            	SCORE -= 10
	        

	    if(not in_track):
	        SCORE -= 100
	    return SCORE


	def coordinateMapping(self, coordinates):
		global GOAL_ONE

		x = coordinates[0]
		y = coordinates[1]

		x = x + GOAL_ONE[0]
		y = y + GOAL_ONE[1]

		#return coordinates

		return (int(x),int(y))

	def move_to_car(self, coordinates):
		global GOAL_ONE

		x = coordinates[0]
		y = coordinates[1]

		x = x - GOAL_ONE[0]
		y = y - GOAL_ONE[1]

		#return coordinates

		return (int(x),int(y))

	def move_to_mid(self, coordinates):
		global HEIGHT
		global WIDTh

		x = coordinates[0]
		y = coordinates[1]

		return (int(x + WIDTH/2),int(y + HEIGHT/2))

	def cones_to_car_dist(self, RFID):
		global carPos
		global N_LM
		
		sorted = []
		for i in RFID:
			sorted.append(self.dist((i[0],i[1]), (carPos.x, carPos.y)))

		return sorted

	def reset(self):
		global VELOCITY
		global DT
		global mids
		global GOAL_ONE
		global RFID
		global carPos
		global N_LM
		global SCORE
		global LEVEL


		# State Vector [x y yaw v]'
		#xEst = np.array([[mids[0][0]], [mids[0][1]], [0]])  # SLAM estimation
		self.xEst = np.zeros((STATE_SIZE, 1))  # True state
		#xTrue = np.array([[mids[0][0]], [mids[0][1]], [0]])
		self.xTrue = np.zeros((STATE_SIZE, 1))
		#xDR = np.array([[mids[0][0]], [mids[0][1]], [0]])  # Dead reckoning
		self.xDR = np.zeros((STATE_SIZE, 1))

		# history
		self.hxEst = self.xEst
		self.hxTrue = self.xTrue
		self.hxDR = self.xTrue

		self.particles = [Particle(N_LM) for i in range(N_PARTICLE)]

		SCORE = 0
		LEVEL = 0



	def state(self):
		global RFID
		global VELOCITY
		global STEERING
		global ACCELERATION
		global carPos
		global SCORE

		state = np.zeros(shape = (500,2))

		car_pos = np.array([carPos[0], carPos[1]])
		velocity = np.array([VELOCITY.x,VELOCITY.y])
		other_shit = np.array([STEERING,ACCELERATION])

		#sort landmarks by distance to car
		distances = self.cones_to_car_dist(RFID)
		order = np.argsort(distances)
		sorted_cones = RFID[order]

		state_temp = np.vstack((car_pos,velocity,other_shit, sorted_cones))
		state[:state_temp.shape[0],:state_temp.shape[1]] = state_temp

		return state


		#https://keon.io/deep-q-learning/



