'''Created by David Buchanan; both original algorithm and code'''
import random
from math import sin, cos, pi, sqrt, atan2
import os
tau = 2*pi

TRACK_WIDTH = 30
MIN_RADIUS = 12
MAX_RADIUS = 50
MIN_ANGLE = 0.2*pi
MAX_ANGLE = 1.5*pi
STEP_SIZE = 5

CONE_DIST = 30
MAP_RADIUS = 400
START_POINT = (0, 0)

def interp(x0, y0, x1, y1, n):
	x = x0*n + x1*(1-n)
	y = y0*n + y1*(1-n)
	return x, y, 0

def fixup_angles(points):
	for a, b, i in zip(points, points[1:]+[points[0]], range(len(points))):
		points[i] = (a[0], a[1], atan2(a[0]-b[0], a[1]-b[1]))

def getedge(points, angle):
	edgepoints = []
	TR = TRACK_WIDTH/2
	for x, y, r in points:
		norm = r+angle
		edgepoints.append((x+sin(norm)*TR, y+cos(norm)*TR, 0))
	return edgepoints

def calc_cones(points):
	lastdist = CONE_DIST
	conepoints = []
	for a, b in zip(points[:-1], points[1:]):
		x0, y0, _ = a
		x1, y1, _ = b
		dist = sqrt(dist2(x0, y0, x1, y1))
		lastdist += dist
		if lastdist > CONE_DIST:
			lastdist -= CONE_DIST
			overshoot = lastdist/dist
			conepoints.append(interp(x0, y0, x1, y1, overshoot))
	return conepoints

def dist2(x0, y0, x1, y1):
	return (x0-x1)*(x0-x1)+(y0-y1)*(y0-y1)

def xy(points):
	return [(x, y) for x, y, *_ in points]

def end_selfintersects(points, ending=True, width=TRACK_WIDTH):
	endx, endy, _ = points[-1]
	TW2 = width*width
	if ending:
		if dist2(endx, endy, 0, 0) > MAP_RADIUS*MAP_RADIUS:
			return True
	for x, y, _ in reversed(points[:-2]):
		if ending:
			if dist2(endx, endy, x, y) > TW2:
				ending = False
		else:
			if dist2(endx, endy, x, y) < TW2:
				return True
	return False

def append_turn(midpoints, radius, radians):
	x, y, r = midpoints[-1]
	turn_radius = 50
	arclength = abs(radians*radius)
	numsteps = int(arclength/STEP_SIZE)
	if numsteps == 0:
		return
	stepsize = (arclength/numsteps)
	rstepsize = (radians/numsteps)
	for i in range(numsteps):
		x += sin(r)*stepsize
		y += cos(r)*stepsize
		r += rstepsize
		midpoints.append((x, y, r))
		if end_selfintersects(midpoints):
			return False
	return True
	


def gen_track():
	while True:
		midpoints = [ # x, y, radians
			(START_POINT[0], START_POINT[1], 0),
		]

		num_failures = 0
		num_steps = 0
		done = False

		append_turn(midpoints, 10000, 0.01)
		START_LEN = len(midpoints)

		# run the game loop
		while not done:

			if len(midpoints) > 150:
				lastpoint = midpoints[-1]
				if dist2(lastpoint[0], lastpoint[1], START_POINT[0], START_POINT[1]) < 40*40:
					
					CROSSOVER = 10
					firstpoint = midpoints[0]
					dist = sqrt(dist2(firstpoint[0], firstpoint[1], lastpoint[0], lastpoint[1]))
					num_steps = int(dist/STEP_SIZE)
					bridge = midpoints[-CROSSOVER:]
					for i in range(1, num_steps):
						bridge.append(interp(firstpoint[0], firstpoint[1], lastpoint[0], lastpoint[1], i/num_steps))
					bridge += midpoints[:CROSSOVER]
					
					
					NSTEPS = 10
					for j in range(NSTEPS):
						indent = 1+int(len(bridge)/2 * j/NSTEPS)
						for i in range(indent, len(bridge)-indent):
							prev, next = bridge[i-1], bridge[i+1]
							midpoint = interp(prev[0], prev[1], next[0], next[1], 0.5)
							bridge[i] = interp(bridge[i][0], bridge[i][1], midpoint[0], midpoint[1], 1/NSTEPS)
					
					midpoints = bridge + midpoints[CROSSOVER:-CROSSOVER]
					
					for i in range(2, len(midpoints)-10):
						if end_selfintersects(midpoints[:i]):
							done = True
							break
					if done:
						break
					
					fixup_angles(midpoints)
					midpoints.append(midpoints[0])
					#print([x[2] for x in midpoints])
					
					xs = [x[0] for x in midpoints]
					ys = [x[1] for x in midpoints]
					midx = (min(xs)+max(xs))/2
					midy = (min(ys)+max(ys))/2
					
					#midpoints = [(x-midx, y-midy, r) for x, y, r in midpoints]
					
					yellowcones = calc_cones(getedge(midpoints, pi/2))
					bluecones = calc_cones(getedge(midpoints, -pi/2))
					for pos in yellowcones+bluecones:
						if end_selfintersects(midpoints+[pos], False, TRACK_WIDTH*0.5*0.8):
							done = True
							break
					if done:
						break
					
					
					return xy(yellowcones), xy(bluecones), midpoints
			
			old = midpoints.copy()
			if random.random() > 0.7:
				ok = append_turn(midpoints, random.uniform(MIN_RADIUS, MAX_RADIUS), random.uniform(-MAX_ANGLE, MAX_ANGLE))
			else:
				ok = append_turn(midpoints, 1000, random.uniform(pi*0.005, pi*0.02))
			if not ok:
				num_failures += 1
				midpoints = old
				if num_failures > 10:
					num_failures = 0
					midpoints = midpoints[:START_LEN]+midpoints[START_LEN:-20]
			else:
				num_failures = 0
			
			num_steps += 1
			if num_steps > 1000:
				break
